import os
import cv2
from .ffmpeg import DmFFMPEG

dmffmpeg = DmFFMPEG()


def push_stream(
        source,
        ip='127.0.0.1',
        port=1935,
        stream_type='rtmp',
        key=None,
        vcodec='h264',
        acodec=None,
        suffix='.jpg',
        **kwargs, ):
    """
    Push stream via DmFFMPEG based on FFMPEG.
    Note: You need configure Nginx to serve the stream.

    :param source: The source needs to be pushed. It could be single image (numpy.array), folder contains images, stream (rtsp or rtmp) or video (.mp4, .avi etc.), .
    :type source: str
    :param ip: The ip of Nginx. Default is "127.0.0.1".
    :type ip: str
    :param port: The port of Nginx. Default is 1935.
    :type port: int
    :param stream_type: The stream type, "rtmp" or "rtsp". Default is "rtmp".
    :type stream_type: str
    :param key: The key of the stream, which is the additional chars in url. Default is None.
    :type key: str
    :param vcodec: The video codec for ffmpeg. Default is "h264".
    :type vcodec: str
    :param acodec: The audio codec for ffmpeg. Default is None.
    :type acodec: str
    :param suffix: The suffix of the image, valid only when source is a folder. Default is ".jpg".
    :type suffix: str
    :param kwargs: Other parameters for ffmpeg.
    :return: None, The stream will be pushed to Nginx server in url: stream_type://ip:port/live/key, i.e. rtmp://127.0.0.1:1935/live.

    Example:
        >>> import damei as dm
        >>> dm.ffmpeg.push_stream(source="your numpy.array, video, rtsp or rtmp stream etc.")

        >>> $ ffplay "rtmp://127.0.0.1:1935/live"  # to show the stream

    Example2 Push stream from video:
        >>> import damei as dm
        >>> video_path = "your video path"
        >>> dm.ffmpeg.push_stream(source=video_path)

    Eaxmple3 Push stream from images (np.array):
        >>> import damei as dm
        >>> import cv2
        >>> img_files = ["your image path1", "your image path2"]
        >>> for i, img_file in enumerate(img_files):
        >>>     img = cv2.imread(img_file)  # read image from file, and convert to numpy.array
        >>>     dm.ffmpeg.push_stream(img)
    """

    if isinstance(source, str) and os.path.isdir(source):
        img_files = [f'{source}/{x}' for x in os.listdir(source) if x.endswith(suffix)]
        for i, img_file in enumerate(img_files):
            assert os.path.exists(img_file), f'Could Not Found Image: {img_file}'
            img = cv2.imread(img_file)
            dmffmpeg.push_stream(
                source=img,
                ip=ip,
                port=port,
                stream_type=stream_type,
                key=key,
                vcodec=vcodec,
                acodec=acodec,
                **kwargs,
            )

    else:
        dmffmpeg.push_stream(
            source=source,
            ip=ip,
            port=port,
            stream_type=stream_type,
            key=key,
            vcodec=vcodec,
            acodec=acodec,
            **kwargs,
        )



"""
dm关于时间的函数
"""
import datetime


def current_time(format='%Y-%m-%d %H:%M:%S'):
    """
    读取当前时间

    :param format: format
    :return: time, str

    Example:
        >>> ctime = dm.current_time()
    """
    return datetime.datetime.now().strftime(format)


def plus_time(ctime=None, seconds=1, format='%Y-%m-%d %H:%M:%S'):
    """
    添加时间

    :param ctime: current time, str
    :param seconds: seconds, int
    :param format: format, str
    :return: time, str

    Example:
        >>> future_time = dm.plus_time(ctime=ctime, seconds=50*24*60*60)
    """
    ct = ctime if ctime else current_time(format=format)
    ct = datetime.datetime.strptime(ct, format)
    return (ct + datetime.timedelta(seconds=seconds)).strftime(format)


def within_time(expiration_time, ctime=None, format='%Y-%m-%d %H:%M:%S'):
    """
    输入到期时间，判断当前时间是否在到期时效内

    :param expiration_time: 到期时间, str
    :param ctime: 当前时间，str. Default: current_time()
    :param format: 格式，str
    :return: True or False, bool

    Example:
        >>> is_within = dm.within_time(expiration_time=future_time, current_t=None)
    """
    ct = ctime if ctime else current_time(format=format)  # 读取当前时间，str格式
    # ct = plus_time(ct, seconds=50*24*60*60)  # 测试用的，
    ct = datetime.datetime.strptime(ct, format)  # 转为date格式
    et = datetime.datetime.strptime(expiration_time, format)  # 转为date格式
    return ct < et



from damei.misc.logger import getLogger

logger = getLogger('exception.py')


def exception_handler(error_type=None, error_info=None, **kwargs):
    # error = kwargs.get('error', )
    mute = kwargs.get('mute', False)
    if mute:
        return
    need_traceback = kwargs.get('traceback', False)
    info2 = kwargs.get('info', None)

    info = f'{error_type}' if error_type else ''
    # print(info)
    info += f': {error_info}.' if error_info else info
    # print(info)
    info += f' {info2}' if info2 else info
    # print(info)

    if need_traceback:
        info += f'\n{error_info.__traceback__.tb_frame.f_globals["__file__"]}:{error_info.__traceback__.tb_lineno}'
    logger.info(info)



#!/home/zzd/anaconda3/envs/yolov5/bin/python
# coding=utf-8
"""
###!/Anaconda3/envs/yolov5/bin/python
damei scp 配置好目标机器，同步文件
usage:
    # 上传
    dmscp siyuan -u xxx.zip --rdir /path/to/remote_server_dir
    dmscp siyuan -d ./xxx.zip -d # 下载 -d --download
    dmscp server list  #
    dmscp config_server
"""
import argparse
import os
from pathlib import Path

import paramiko

import damei as dm

logger = dm.getLogger('dmscp', name_lenth=5)


class RemoteServer(object):
    """远程服务器"""

    def __init__(self, name, host, user, passwd=None, port=22, key_file=None):
        self.name = name
        self.host = host
        self.port = port
        self.user = user
        self.passwd = passwd
        self.key_file = key_file
        self.ssh = None

    def __repr__(self):
        passwd = '*********' if self.passwd else 'No_passwd'
        key_file = self.key_file if self.key_file else ''
        format_str = f'{self.user}@{self.host}:{self.port} {passwd} {key_file}'
        return format_str

    def _init_ssh(self):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh = ssh

    def exec(self, cmd):
        assert self.ssh is not None
        stdin, stdout, stderr = self.ssh.exec_command(cmd)
        res, err = stdout.read(), stderr.read()
        result = res if res else err
        # print(result.decode('utf-8'))
        return result.decode('utf-8')

    def connect(self):
        """连接远程服务器"""
        if self.ssh is None:
            self._init_ssh()
        try:
            self.ssh.connect(self.host, self.port, self.user, self.passwd)
        except Exception as e:
            logger.error(f'SSH connect to server {self.user}@{self.host} port: {self.port} failed：{e}')
            # print(self.passwd)
            return False
        return True

    def disconnect(self):
        """断开远程服务器"""
        self.ssh.close()


class DmSCP(object):
    def __init__(self):
        self._default_server = None
        self._servers = self._init_server()

    def _init_server(self):
        home = Path.home()
        rc_file = f'{home}/.damei/.dmscprc'
        assert os.path.exists(rc_file), f'{rc_file} not exists'
        data_dict = dm.misc.rcfile2dict(rc_file)  # 读取rcfile

        _servers = {}

        for k, v in data_dict.items():
            if k == 'default':
                self._default_server = v
            else:
                passwd = v.get('passwd', None)
                port = v.get('port', 22)
                key_file = v.get('key_file', None)
                _servers[k] = RemoteServer(k, v['host'], v['user'], passwd=passwd, port=port, key_file=key_file)
        if self._default_server is None:  # 还是None
            self._default_server = list(_servers.keys())[0]  # 默认第一个
        return _servers

    def __call__(self, *args, **kwargs):
        # print(args, type(args))
        print('args:', args, kwargs)
        first_arg = args[0] if len(args) > 0 else None
        if first_arg == 'servers':
            info = dm.misc.dict2info(self.servers)
            print(info)
        elif first_arg == 'server_names':
            print(self.server_names)
        else:  # 默认上传
            # upload = kwargs.get('upload', False)
            # download = kwargs.get('download', False)

            # 第一个参数是上传文件时
            # server_name = first_arg if first_arg else self.default_server
            if first_arg in self.server_names:
                server_name = first_arg
            else:
                assert not upload, f'{first_arg} not in server_names'
                server_name = self.default_server
                # upload = first_arg
                kwargs['upload'] = first_arg

            # print('xxx', first_arg, args, kwargs, upload)
            if kwargs.get('upload', False):
                self.up_down(server_name, **kwargs)
            elif download:
                self.up_down(server_name, **kwargs)
            else:
                logger.info(f'No upload or download, -u (default) or -d to specify')
                # help = dm.popen(f'dmscp {server_name} --help')
            # print(server_name)
        pass

    @property
    def servers(self):
        return self._servers

    @property
    def server_names(self):
        names = list(self._servers.keys())
        return names

    @property
    def default_server(self):
        return self._default_server
        # return list(self._servers.keys())[0]
        # return self._servers['siyuan']

    def up_down(self, server_name, *args, **kwargs):
        """
        上传或下载
        :param server_name:
        :param source: 文件或文件夹
        """
        upload = kwargs.get('upload', None)
        download = kwargs.get('download', None)
        recursive = kwargs.get('recursive', False)
        force = kwargs.get('force', False)
        remote_dir = kwargs.get('remote_dir', None)

        print(upload, download, recursive, force, remote_dir)
        assert not (upload and download), '-u and -d can not be used together'
        assert upload or download, '-u or -d must be used'
        source = upload if upload else download
        up_or_down = 'upload' if upload else 'download'

        # 获取服务器
        s = self.servers[server_name]
        # server.connect()  # 暂时不需要连接

        # 处理source
        src = Path(os.path.abspath(source))
        # 处理递归参数
        recursive = '-r ' if recursive else ''
        # 处理remote dir
        local_home = f'{Path(os.path.expanduser("~"))}'
        local_dir = src.parent
        remote_dir = remote_dir if remote_dir else f'{src.parent}'
        remote_dir = remote_dir.replace(local_home, '~') if remote_dir else None  # 必须是~/xx的形式

        port = s.port
        # code = f'scp {recursive}-P {port} {src} {s.user}@{s.host}:~{rel_dir}/{src.name}'
        if upload:
            code = f'scp {recursive}-P {port} {src} {s.user}@{s.host}:{remote_dir}/{src.name}'
        else:  # download
            code = f'scp {recursive}-P {port} {s.user}@{s.host}:{remote_dir}/{src.name} {src}'

        # print(code)
        logger.info(code)
        # print(src_dir)
        # print(server_name, 'xx')
        # exec(code)
        ret = dm.popen(code)  # 传输成功返回[]，失败可能是：['scp: xxx/misc.py: No such file or directory']

        if len(ret) > 0:
            # logger.warn(f'Upload failed: {ret[0]}')
            if force:
                logger.info(f'Force {up_or_down} ...')
                # 连接远程服务器，创建目录，重新上传
                is_connected = s.connect()
                if is_connected:
                    s.exec(f'mkdir -p {remote_dir}')
                    self.up_down(server_name, *args, **kwargs)
            else:
                logger.error(f'{up_or_down} failed: {ret[0]}, you may use -f to force {up_or_down}.')
                # raise ValueError(f'Upload failed: {ret}')
        else:
            logger.info(f'{up_or_down} {source} success.')

    def download(self, src, dst):
        pass

    def config_server(self, server):
        pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='damei scp')

    parser.add_argument('args', nargs='*', default=None, help='servers,  ')  # 或多个参数
    parser.add_argument('-u', '--upload', type=str, default=None, help='upload')
    parser.add_argument('-d', '--download', type=str, default=None, help='download')
    parser.add_argument('-r', '--recursive', action='store_true', default=False, help='recursive copy all sub files')
    parser.add_argument('-f', '--force', action='store_true', default=False,
                        help='force to copy by makedirs in remote if the remote dir not exist')
    parser.add_argument('-rdir', '--remote_dir', type=str, default=None,
                        help='remote dir for save file, default is the same as local dir')

    opt = parser.parse_args()
    print(opt)
    dmscp = DmSCP()

    args = opt.args
    upload = opt.upload
    download = opt.download

    dmscp(*args, upload=upload, download=download, recursive=opt.recursive,
          force=opt.force, remote_dir=opt.remote_dir,
          )



from damei.nn.uaii.stream.base_module import AbstractModule
from damei.nn.uaii.stream.base_input import AbstractInput
from damei.nn.uaii.stream.base_output import AbstractOutput
from damei.nn.uaii.stream.base_queue import AbstractQue



"""
利用ffmpeg推流
"""
import os
import subprocess

import numpy as np

from damei.misc.logger import getLogger

logger = getLogger('DmFFMPEG')


class DmFFMPEG(object):
    def __init__(
            self,
            ip='127.0.0.1',
            port='1935',
            stream_type='rtmp',
            key=None,
            vcodec='h264',
            acodec=None,
            **kwargs
    ):
        self.ip = ip
        self.port = port
        self.stream_type = stream_type
        self.key = key
        self.vcodec = vcodec
        self.acodec = acodec
        self._pipe = None
        self.mute = kwargs.pop('mute', False)  #

    def init_pipe(self, img, **kwargs):
        """初始化管道，第一次调用push_stream时且传入为图像时，调用"""
        ip = kwargs.pop('ip', self.ip)
        port = kwargs.pop('port', self.port)
        stream_type = kwargs.pop('stream_type', self.stream_type)
        key = kwargs.pop('key', self.key)
        vcodec = kwargs.pop('vcodec', self.vcodec)
        acodec = kwargs.pop('acodec', self.acodec)
        fps = kwargs.pop('fps', 15)

        key = f'/{key}' if key else ''
        vcodec = f'{vcodec}' if vcodec else ''
        acodec = f'{acodec}' if acodec else ''

        url = f'{stream_type}://{ip}:{port}/live{key}'
        h, w, c = img.shape
        command = [
            'ffmpeg',
            # 're',#
            # '-y', # 无需询问即可覆盖输出文件
            '-f', 'rawvideo',  # 强制输入或输出文件格式
            # '-vcodec','rawvideo', # 设置视频编解码器。这是-codec:v的别名
            '-pix_fmt', 'bgr24',  # 设置像素格式
            '-s', f'{w}x{h}',  # 设置图像大小
            '-r', f'{fps}',  # 设置帧率
            '-i', '-',  # 输入
            '-vcodec', vcodec,
            '-acodec', acodec,
            # '-pix_fmt', 'yuv420p',
            # '-preset', 'ultrafast',
            '-f', 'flv',  # 强制输入或输出文件格式
            url]
        # print(command)
        # exit()
        logger.info(f'The stream will be pushed to: {url}.')

        if self.mute:
            pipe = subprocess.Popen(
                command, stdin=subprocess.PIPE, stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE)
        else:
            pipe = subprocess.Popen(
                command, stdin=subprocess.PIPE,
            )
        # exit()
        self._pipe = pipe
        return pipe

    def get_pipe(self, img=None, **kwargs):
        if self._pipe is not None:
            return self._pipe
        else:
            return self.init_pipe(img=img, **kwargs)

    # return self._pipe

    def push_stream(self, source, **kwargs):
        """
        推流
        :param source: 可以是rtsp rtmp流或.mp4 .avi等视频文件
                        也可以是单张图，必须是np.ndarray格式
        :param kwargs:
        :return:
        """
        ip = kwargs.pop('ip', self.ip)
        port = kwargs.pop('port', self.port)
        stream_type = kwargs.pop('stream_type', self.stream_type)
        key = kwargs.pop('key', self.stream_type)
        vcodec = kwargs.pop('vcodec', self.vcodec)
        acodec = kwargs.pop('acodec', self.acodec)
        key = f'/{key}' if key else ''
        vcodec = f'-vcodec {vcodec}' if vcodec else ''
        acodec = f'-acodec {acodec}' if acodec else ''

        # 推视频流
        if isinstance(source, str):
            code = f'ffmpeg -i {source} {vcodec} {acodec} -f flv {stream_type}://{ip}:{port}/live{key}'
            if self.mute:
                subprocess.Popen(code, stdout=subprocess.STDOUT, stderr=subprocess.STDOUT)
            else:
                os.system(code)
        elif isinstance(source, np.ndarray):
            # 第一次调用时检查是否有输出pipe
            pipe = self.get_pipe(source, **kwargs)
            pipe.stdin.write(source.tobytes())
        else:
            raise NotImplementedError(f'Source type error: {source} {type(source)}.')

    def show_stream(self, source=None, stream_type='rtmp', ip='127.0.0.1', port='1935', key=None):
        key = f'/{key}' if key else ''
        source = source if source else f'{stream_type}://{ip}:{port}/live{key}'
        code = f'ffplay {source}'
        os.system(code)



"""
damei nn api for ai algorithm management (unified ai interface)
"""

from .base import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
from .registry import MODULES, SCRIPTS, IOS
from .utils import Config
from damei.nn.uaii.uaii_main import UAII

# pass



import logging
import os
from pathlib import Path


def get_logger(name=None, **kwargs):
    return getLogger(name, **kwargs)


def getLogger(name=None, **kwargs):
    name = name if name else 'root'
    name_lenth = kwargs.get('name_lenth', 12)
    name = f'{name:<{name_lenth}}'
    logger = logging.getLogger(name)
    level = kwargs.get('level', logging.INFO)
    format_str = f"\033[1;35m[%(asctime)s]\033[0m \033[1;32m[%(name)s]\033[0m " \
                 f"\033[1;36m[%(levelname)s]:\033[0m %(message)s"
    logging.basicConfig(level=level,
                        format=format_str,
                        # datefmt='%d %b %Y %H:%M:%S'
                        )
    logg_dir = f'{Path.home()}/.damei/logs/{Path(os.getcwd()).name}'
    if not os.path.exists(logg_dir):
        os.makedirs(logg_dir)
    fh = logging.FileHandler(f'{logg_dir}/{Path(os.getcwd()).name}.log')
    fh.setLevel(level=level)
    # ch = logging.StreamHandler()
    # ch.setLevel(level=level)
    fh.setFormatter(logging.Formatter(format_str))
    # ch.setFormatter(logging.Formatter(format_str))
    logger.addHandler(fh)
    # logger.addHandler(ch)
    return logger



import base64


class DmRsa(object):
    """这是加密Encrypt方案"""

    def __init__(self, plan='Encrypt'):
        """
        :param plan: 加密方案Encrypt或签密方案SignEncripy
        """
        # import rsa as raw_rsa
        self._raw_rsa = None
        self.length = 1024

    @property
    def raw_rsa(self):
        if not self._raw_rsa:
            import rsa as raw_rsa
            self._raw_rsa = raw_rsa
        return self._raw_rsa

    def gen_rsa(self, num=1, length=None):
        """在随机生成公钥-私钥对，存储在当前路径pubk.txt和privk.txt里"""
        length = length if length else self.length
        # rsa_length = int(length / 8)
        # enable_length = rsa_length - 11
        for i in range(num):
            pubk, privk = self.gen_pair(length=length)
            self._save(pubk=pubk, privk=privk, suffix=f'{i + 1}')

    def gen_pair(self, length):
        pub_key_obj, priv_key_obj = self.raw_rsa.newkeys(length)
        pub_key_str = pub_key_obj.save_pkcs1()
        pub_key_code = base64.standard_b64encode(pub_key_str).decode()

        priv_key_str = priv_key_obj.save_pkcs1()
        priv_key_code = base64.standard_b64encode(priv_key_str).decode()
        return pub_key_code, priv_key_code

    def _save(self, pubk, privk, suffix=''):
        with open(f'pubk{suffix}.pem', 'w') as f:
            f.write(pubk)
        with open(f'privk{suffix}.pem', 'w') as f:
            f.write(privk)

    def load_pk(self, pubk=None, privk=None):
        """加载钥匙"""
        assert pubk or privk  # 至少有一个
        assert not (pubk and privk)  # 不能同时都有

        if (pubk is not None) and (privk is not None):
            # print(privk, pubk, pubk is not None, privk is not None)
            raise KeyError(f'The input public key and private key both not None, the software dont know which to load.')
        elif pubk:
            if pubk.endswith('.pem'):
                with open(pubk, 'r') as f:
                    pubk = f.read()
            pubk = base64.standard_b64decode(pubk)  # 对公钥解密
            pubk = self.raw_rsa.PublicKey.load_pkcs1(pubk)  # 加载公钥
            # print(f'pubk2: {pubk}')
            return pubk
        elif privk:
            if privk.endswith('.pem'):
                with open(privk, 'r') as f:
                    privk = f.read()
            privk = base64.standard_b64decode(privk)  # 私钥解密
            privk = self.raw_rsa.PrivateKey.load_pkcs1(privk)  # 加载私钥
            return privk
        else:
            raise KeyError(f'Both public key and private key are None.')

    def encrypt(self, plaintext, pubk=None, privk=None, length=None):
        """
        用公钥或私钥对明文进行rsa加密 + base64加密
        :param plaintext，明文，str类型
        :param pubk: 公钥，str类型，需先解密，直接是公钥，或者是以.txt结尾的公钥文件
        :return: 密文，str类型
        """
        length = length if length else self.length
        # rsa加密
        pk = self.load_pk(pubk=pubk, privk=privk)
        rsa_length = int(length / 8)
        enable_length = rsa_length - 11
        ciphertext_list = []
        for i in range(0, len(plaintext), enable_length):
            v = plaintext[i:i + enable_length]
            val = self.raw_rsa.encrypt(v.encode(), pk)  # 加密，加密后是bytes类型
            ciphertext_list.append(val)
        ciphertext = b''.join(ciphertext_list)

        # base64加密
        ciphertext = base64.b64encode(ciphertext)
        ciphertext = ciphertext.decode('utf-8')

        return ciphertext

    def decrypt(self, ciphertext, privk=None, pubk=None, length=None):
        """
        利用私钥或公钥，对密文进行解密，对应的rsa+base64加密，解密是base64+rsa解密
        :param ciphertext: 密文，str类型
        :param privk: 私钥，str类型
        :param length: 1024
        :return: 明文，str类
        """
        # 先base64解密
        ciphertext = base64.b64decode(ciphertext.encode())

        # rsa解密
        length = length if length else self.length
        pk = self.load_pk(pubk=pubk, privk=privk)
        # print(f'pk: {pk} {type(pk)}')
        rsa_length = int(length / 8)
        plaintext_list = []
        for i in range(0, len(ciphertext), rsa_length):
            v = ciphertext[i:i + rsa_length]
            val = self.raw_rsa.decrypt(v, pk)
            plaintext_list.append(val)

        plaintext = b''.join(plaintext_list)
        plaintext = plaintext.decode('utf-8')
        return plaintext

        pass

    def sign(self, message, privk, hash_method='MD5'):
        """
        Sign a message by private key
        :param message: 消息
        :param privk: 私钥
        :param hash_method: MD5, SHA-1, SHA-256, SHA-512等
        :return: 签名，16进制的str
        """
        privk = self.load_pk(privk=privk)
        signature = self.raw_rsa.sign(message.encode('utf-8'), privk, hash_method=hash_method)
        # print('xxx', signature)
        signature = signature.hex()
        return signature

    def verify(self, message, signature, pubk):
        pubk = self.load_pk(pubk)
        # print('xx2', signature)
        # signature = signature.encode('utf-8')
        signature = bytes.fromhex(signature)
        # print('xx3', signature)
        try:
            self.raw_rsa.verify(message.encode('utf-8'), signature, pubk)
            return True
        except Exception as e:
            print(f'Verify failed, {e}')
            return False
# print(ret)
